# -*- coding: utf-8 -*-
import numpy as np

from ..partonic_channel import RSL
from . import partonic_channel as pc


class GluonVV(pc.NeutralCurrentBase):
    def NLO(self):
        """
        |ref| implements :eqref:`D.1`, :cite:`felix-thesis`.
        """

        def cg(z, args):
            if self.is_below_threshold(z):
                return 0.0
            # fmt: off
            return self._FHprefactor * (
                (-np.pi * self._rho_p(z) ** 3)
                / (4 * self._rho(z) ** 2 * self._rho_q ** 2)
                * (
                    2 * self._beta(z) * (
                        self._rho(z) ** 2
                        + self._rho_q ** 2
                        + self._rho(z) * self._rho_q * (6 + self._rho_q)
                    )
                    +
                    np.log(self._chi(z)) * (
                        2 * self._rho_q ** 2 * (1 + self._rho(z))
                        + self._rho(z) ** 2 * (2 - (self._rho_q - 4) * self._rho_q)
                    )
                )) / z
            # fmt: on

        return RSL(cg)


class GluonAA(GluonVV):
    def NLO(self):
        """
        |ref| implements :eqref:`D.4`, :cite:`felix-thesis`.
        """

        VV = super().NLO()

        def cg(z, args):
            if self.is_below_threshold(z):
                return 0.0
            return VV.reg(z, args) + self._FHprefactor * np.pi / 2.0 * (
                self._rho_p(z) * self._rho_q * np.log(self._chi(z))
            )

        return RSL(cg)



import numpy as np

from ..observable_name import ObservableName
from .result import EXSResult

conv = 3.893793e10  # conversion factor from GeV^-2 to 10^-38 cm^2


class EvaluatedCrossSection:
    def __init__(self, xs, kin):
        self.xs = xs
        self.kin = kin

    def f_coeffs(self):
        y = self.kin["y"]
        yp = 1.0 + (1.0 - y) ** 2
        ym = 1.0 - (1.0 - y) ** 2
        yL = y ** 2
        obs_config = self.xs.runner.managers["coupling_constants"].obs_config
        f3sign = -1 if obs_config["projectilePID"] < 0 else 1
        # eta = 1 if obs_config["process"] == "CC" else 1
        # the alpha_qed^2 part is shifted below
        # norm = 2.0 * np.pi / (y * x * Q2)
        kind = self.xs.obs_name.kind
        # if kind == "XSreduced":
        #     return np.array([1.0, -yL / yp, f3sign * ym / yp])
        # if kind == "XSyreduced":
        #     return np.array([yp, -yL, f3sign * ym])
        if kind == "XSHERANCAVG":
            return np.array([1.0, -yL / yp, 0.0])
        if kind == "XSHERANC":
            return np.array([1.0, -yL / yp, f3sign * ym / yp])
        norm = 0.0
        if kind == "XSHERACC":
            norm = 1.0 / 4.0
        else:
            x = self.kin["x"]
            Q2 = self.kin["Q2"]
            mn = np.sqrt(self.xs.runner.theory_params["M2target"])
            m2w = self.xs.runner.theory_params["M2W"]
            yp -= 2.0 * (mn * x * y) ** 2 / Q2  # = ypc
            # Chorus
            if kind == "XSCHORUSCC":
                norm = (
                    conv
                    * self.xs.runner.theory_params["GF"] ** 2
                    * mn
                    / (2.0 * np.pi * (1.0 + Q2 / m2w) ** 2)
                )
            if kind == "XSNUTEVCC":
                norm = 100.0 / 2.0 / (1.0 + Q2 / m2w) ** 2
        return np.array([yp, -yL, f3sign * ym]) * norm

    def alpha_qed_power(self):
        return 0

    def get_result(self):
        # Collect esfs
        flavor = self.xs.obs_name.flavor
        f_coeffs = self.f_coeffs()
        f2 = self.xs.get_esf(ObservableName(f"F2_{flavor}"), self.kin).get_result()
        fl = self.xs.get_esf(ObservableName(f"FL_{flavor}"), self.kin).get_result()
        # skip F3 if it is not required
        if f_coeffs[2] != 0.0:
            f3 = self.xs.get_esf(ObservableName(f"F3_{flavor}"), self.kin).get_result()
        else:
            f3 = np.zeros_like(f2)

        # add normalizations
        esf = f_coeffs @ np.array([f2, fl, f3])
        # remap to EXS
        sigma = EXSResult(self.kin["x"], self.kin["Q2"], self.kin["y"], f2.nf)
        for o, v in esf.orders.items():
            # now shift orders: push alpha_qed two powers up
            sigma.orders[(o[0], o[1] + self.alpha_qed_power(), o[2], o[3])] = v
        return sigma



# -*- coding: utf-8 -*-
import numba as nb
import numpy as np


@nb.njit("f8(f8,f8[:])", cache=True)
def sing_from_distr_coeffs(z, coeffs):
    log_ = np.log(1 - z)
    res = 0
    for k, coeff in enumerate(coeffs):
        res += coeff * 1 / (1 - z) * log_ ** k
    return res


@nb.njit("f8(f8,f8[:])", cache=True)
def loc_from_distr_coeffs(x, coeffs):
    log_ = np.log(1 - x)
    res = 0
    delta = coeffs[0]
    for k, coeff in enumerate(coeffs[1:]):
        res += coeff * log_ ** (k + 1) / (k + 1)
    return res + delta


@nb.njit("f8(f8,f8[:])", cache=True)
def loc_from_delta(_, coeffs):
    return coeffs[0]


class RSL:
    """
    RSL representation of a distribution, containing Dirac delta and plus
    distributions.

    Parameters
    ----------
    reg : callable
        regular part
    sing : callable
        singular part
    loc : callable
        local part
    args : sequence or dict
        arguments to pass to the individual parts (if dict) or to all of them
        (if sequence); if no arguments needed `None` is available

    """

    def __init__(self, reg=None, sing=None, loc=None, args=None):
        self.reg = reg
        self.sing = sing
        self.loc = loc
        if isinstance(args, dict):
            self.args = {
                k: np.array(args[k], dtype=float)
                if k in args and args[k] is not None
                else np.array([], dtype=float)
                for k in ["reg", "sing", "loc"]
            }
        else:
            self.args = {
                k: np.array(args, dtype=float)
                if args is not None
                else np.array([], dtype=float)
                for k in ["reg", "sing", "loc"]
            }

    @classmethod
    def from_distr_coeffs(cls, reg, coeffs, reg_args=None):
        """
        Compute the RSL structure form the coefficients of the distributions.

        Parameters
        ----------
            regular : callable
                regular piece (passed unchanged)
            delta : float
                coefficient of the Dirac-delta function
            coeffs: list(float)
                coefficients of the plus-distributions with increasing power of log

        """
        return cls(
            reg=reg,
            sing=sing_from_distr_coeffs,
            loc=loc_from_distr_coeffs,
            args={"reg": reg_args, "sing": coeffs[1:], "loc": coeffs},
        )

    @classmethod
    def from_delta(cls, delta_coeff):
        return cls(loc=loc_from_delta, args={"loc": [delta_coeff]})

    def __repr__(self):
        r = "r" if self.reg is not None else "-"
        s = "s" if self.sing is not None else "-"
        l = "l" if self.loc is not None else "-"
        return f"RSL({r},{s},{l}) - args: {self.args}"


class PartonicChannel(dict):
    """
    Container of partonic coefficient functions

    Parameters
    ----------
        ESF : yadism.structure_function.esf.EvaluatedStructureFunction
            parent ESF
        nf : int
            number of pure light flavors
    """

    def __init__(self, ESF, nf):
        super().__init__()
        self.ESF = ESF
        self.nf = nf
        # default coeff functions to 0
        self[0] = self.decorator(self.LO)
        self[1] = self.decorator(self.NLO)
        self[2] = self.decorator(self.NNLO)

    def convolution_point(self):
        """
        Convolution point
        """
        return self.ESF.x

    def decorator(self, f):
        """
        Deactivate preprocessing

        Parameters
        ----------
            f : callable
                input

        Returns
        -------
            f : callable
                output
        """
        return f

    @staticmethod
    def LO():
        return None

    @staticmethod
    def NLO():
        return None

    @staticmethod
    def NNLO():
        return None


class EmptyPartonicChannel(PartonicChannel):
    def __init__(self, *args, **_kwargs):
        super().__init__(*args)



# -*- coding: utf-8 -*-

import numbers

import numpy as np


class ESFResult:
    """
    Represents the output tensor for a single kinematic point

    Parameters
    ----------
    x : float
        Bjorken x
    Q2 : float
        virtuality of the exchanged boson
    """

    def __init__(self, x, Q2, nf, orders=None):
        self.x = x
        self.Q2 = Q2
        self.nf = nf
        self.orders = {} if orders is None else orders

    @classmethod
    def from_document(cls, raw):
        """
        Recover element from a raw dictionary

        Parameters
        ----------
            raw : dict
                raw dictionary

        Returns
        -------
            new_output : cls
                object representation
        """
        new_output = cls(raw["x"], raw["Q2"], raw["nf"])
        for e in raw["orders"]:
            new_output.orders[tuple(e["order"])] = (
                np.array(e["values"]),
                np.array(e["errors"]),
            )
        return new_output

    def apply_pdf(self, lhapdf_like, pids, xgrid, alpha_s, alpha_qed, xiR, xiF):
        r"""
        Compute the observable for the given PDF.

        Parameters
        ----------
            lhapdf_like : object
                object that provides an xfxQ2 callable (as `lhapdf <https://lhapdf.hepforge.org/>`_
                and :class:`ekomark.toyLH.toyPDF` do) (and thus is in flavor basis)
            pids : list(int)
                list of pids
            xgrid : list(float)
                interpolation grid
            xiF : float
                factorization scale ration :math:`\mu_F^2 = Q^2 \xi_F^2` - beware the square!

        Returns
        -------
            res : dict
                output dictionary with x, Q2, result and error
        """
        if not isinstance(self.Q2, numbers.Number):
            raise ValueError("Q2 is not set!")

        # factorization scale
        muF2 = self.Q2 * xiF ** 2
        pdfs = np.zeros((len(pids), len(xgrid)))
        for j, pid in enumerate(pids):
            if not lhapdf_like.hasFlavor(pid):
                continue
            pdfs[j] = np.array([lhapdf_like.xfxQ2(pid, z, muF2) / z for z in xgrid])

        # join elements
        res = 0
        err = 0
        # join elements
        a_s = alpha_s(np.sqrt(self.Q2) * xiR) / (4 * np.pi)
        alph_qed = alpha_qed(np.sqrt(self.Q2) * xiR)
        for o, (v, e) in self.orders.items():
            lnF = 1.0 if o[3] == 0 else (np.log((1 / xiF) ** 2)) ** o[3]
            lnR = 1.0 if o[2] == 0 else (np.log((1 / xiR) ** 2)) ** o[2]
            prefactor = (a_s ** o[0]) * (alph_qed ** o[1]) * lnR * lnF
            res += prefactor * np.einsum("aj,aj", v, pdfs)
            err += prefactor * np.einsum("aj,aj", e, pdfs)

        return dict(x=self.x, Q2=self.Q2, result=res, error=err)

    def get_raw(self):
        """
        Returns the raw data ready for serialization.

        Returns
        -------
            out : dict
                output dictionary
        """
        d = dict(x=self.x, Q2=self.Q2, nf=self.nf, orders=[])
        for o, (v, e) in self.orders.items():
            d["orders"].append(
                dict(order=list(o), values=v.tolist(), errors=e.tolist())
            )
        return d

    def __add__(self, other):
        r = ESFResult(self.x, self.Q2, self.nf)
        for o, (v, e) in self.orders.items():
            if o in other.orders:  # add the common stuff
                r.orders[o] = (v + other.orders[o][0], e + other.orders[o][1])
            else:  # add my stuff
                r.orders[o] = (v, e)
        for o, (v, e) in other.orders.items():
            if o in self.orders:
                continue
            # add his stuff
            r.orders[o] = (v, e)
        return r

    def __sub__(self, other):
        return self.__add__(-other)

    def __neg__(self):
        return self.__mul__(-1.0)

    def __mul__(self, other):
        r = ESFResult(self.x, self.Q2, self.nf)
        try:
            val = other[0]
            err = other[1]
        except TypeError:
            val = other
            err = 0
        for o, (v, e) in self.orders.items():
            r.orders[o] = (val * v, val * e + err * v)
        return r

    def __rmul__(self, other):
        return self.__mul__(other)


class EXSResult(ESFResult):
    def __init__(self, x, Q2, y, nf, orders=None):
        super().__init__(x, Q2, nf, orders)
        self.y = y

    @classmethod
    def from_document(cls, raw):
        sup = super().from_document(raw)
        return cls(sup.x, sup.Q2, raw["y"], sup.nf, sup.orders)

    def get_raw(self):
        d = super().get_raw()
        d["y"] = self.y
        return d



# -*- coding: utf-8 -*-
import numpy as np

from . import fonll, heavy, intrinsic, kernels, light
from .partonic_channel import EmptyPartonicChannel


class Combiner:
    """
    Does the matching between coefficient functions and partons with their approptiate coupling
    strength.

    Parameters
    ----------
        esf : EvaluateStructureFunction
            current ESF
    """

    def __init__(self, esf):
        self.esf = esf
        self.obs_name = esf.sf.obs_name
        self.nf = esf.sf.threshold.nf(
            esf.Q2
        )  # TODO decide whether Q2 or muF2 is the correct thing
        self.target = esf.sf.target

    def collect_ffns(self):
        """
        Collects all kernels in the |FFNS|.

        Returns
        -------
            elems : list(yadism.kernels.Kernel)
                all participants
        """
        elems = []
        # light is *everything* up to nf and not only u+d+s
        if self.obs_name.flavor in ["light", "total"]:
            elems.extend(light.kernels.generate(self.esf, self.nf))
        if self.obs_name.flavor_family in ["heavy", "total"]:
            if self.obs_name.flavor_family == "heavy":
                # F2b is not avaible in FFNS3
                if self.obs_name.hqnumber > self.nf + 1:
                    raise ValueError(
                        f"{self.obs_name} is not available in FFNS with nl={self.nf}"
                    )
                # F2c in FFNS5 is avaible, but light
                if self.obs_name.hqnumber < self.nf + 1:
                    elems.extend(
                        kernels.generate_single_flavor_light(
                            self.esf, self.nf, self.obs_name.hqnumber
                        )
                    )
                    return elems
                # F2c in FFNS3 is done next
            elems.extend(heavy.kernels.generate(self.esf, self.nf))
            ihq = self.nf + 1
            if ihq in self.esf.sf.intrinsic_range:
                elems.extend(intrinsic.kernels.generate(self.esf, ihq))
        return elems

    def collect_zmvfns(self):
        """
        Collects all kernels for |ZM-VFNS|.

        Returns
        -------
            elems : list(yadism.kernels.Kernel)
                all participants
        """
        elems = []
        # light is *everything* up to nf and not only u+d+s
        if self.obs_name.flavor in ["light", "total"]:
            elems.extend(light.kernels.generate(self.esf, self.nf))
        # heavy is allowed if we already passed it
        if self.obs_name.flavor_family in ["heavy"]:
            if self.obs_name.hqnumber > self.nf:
                raise ValueError(
                    f"{self.obs_name} is not available in ZM-VFNS, yet: nf={self.nf}"
                )
            elems.extend(
                kernels.generate_single_flavor_light(
                    self.esf, self.nf, self.obs_name.hqnumber
                )
            )
        return elems

    def collect_fonll(self):
        """
        Collects all kernels for FONLL.

        Returns
        -------
            elems : list(yadism.kernels.Kernel)
                all participants
        """
        elems = []
        # above the *next* threshold use ZM-VFNS
        nl = self.esf.sf.nf_ff - 1
        if nl + 1 < self.nf:
            return self.collect_zmvfns()

        if self.obs_name.flavor in ["light", "total"]:
            elems.extend(light.kernels.generate(self.esf, nl))
            # add F^d
            elems.extend(
                self.damp_elems(nl, fonll.kernels.generate_light_diff(self.esf, nl))
            )
        if self.obs_name.flavor_family in ["heavy", "total"]:
            elems.extend(heavy.kernels.generate(self.esf, nl))
            # add F^d
            ihq = nl + 1
            if ihq in self.esf.sf.intrinsic_range:
                elems.extend(intrinsic.kernels.generate(self.esf, ihq))
                elems.extend(
                    self.damp_elems(
                        nl,
                        fonll.kernels.generate_heavy_intrinsic_diff(self.esf, nl),
                    )
                )
            else:
                elems.extend(
                    self.damp_elems(nl, fonll.kernels.generate_heavy_diff(self.esf, nl))
                )
        return elems

    def damp_elems(self, nl, elems):
        """
        Damp FONLL difference contributions if necessary.

        Parameters
        ----------
            nl : int
                number of *light* flavors
            elems : list(Kernel)
                kernels to be modified

        Returns
        -------
            elems : list(Kernel)
                modified kernels
        """
        if not self.esf.sf.FONLL_damping:
            return elems
        nhq = nl + 1
        m2hq = self.esf.sf.m2hq[nhq - 4]
        power = self.esf.sf.damping_powers[nhq - 4]
        if self.esf.Q2 > m2hq:
            damp = np.power(1.0 - m2hq / self.esf.Q2, power)
        else:
            damp = 0.0
        return (damp * e for e in elems)

    @staticmethod
    def apply_isospin(full, z, a):
        """
        Apply isospin symmetry to u and d distributions.

        Parameters
        ----------
            full : list(yadism.kernels.Kernel)
                all participants
            z : float
                number of protons
            a : float
                atomic mass number
        """
        nucl_factors = np.array([[z, a - z], [a - z, z]]) / a
        for ker in full:
            for sign in [-1, 1]:
                ps = np.array(
                    [ker.partons.get(sign * 1, 0), ker.partons.get(sign * 2, 0)]
                )
                ker.partons[sign * 1], ker.partons[sign * 2] = nucl_factors @ ps

    @staticmethod
    def drop_empty(full):
        """
        Drop kernels with :class:`EmptyPartonicChannel` or its partons with empty weight.

        Parameters
        ----------
            elems : list(yadism.kernels.Kernel)
                all participants

        Returns
        -------
            filtered_kernels : list(yadism.kernels.Kernel)
                improved participants
        """
        filtered_kernels = []
        for ker in full:
            ker.partons = {p: w for p, w in ker.partons.items() if w != 0}
            if len(ker.partons) > 0 and not isinstance(ker.coeff, EmptyPartonicChannel):
                filtered_kernels.append(ker)
        return filtered_kernels

    def collect_elems(self):
        """
        Collects all kernels according to the |FNS|.

        Returns
        -------
            elems : list(yadism.kernels.Kernel)
                all participants
        """
        if self.esf.sf.scheme == "FFNS":
            full = self.collect_ffns()
        elif self.esf.sf.scheme == "ZM-VFNS":
            full = self.collect_zmvfns()
        elif self.esf.sf.scheme in ["FONLL-A"]:
            full = self.collect_fonll()
        else:
            raise ValueError("Unknown FNS")

        # add level-0 nuclear correction: apply isospin symmetry
        self.apply_isospin(full, self.target["Z"], self.target["A"])

        # drop all kernels with 0 weight, or empty coeffs
        return self.drop_empty(full)



import logging

from .esf import exs

logger = logging.getLogger(__name__)


class CrossSection:
    """
    Represents an abstract structure function.

    This class acts as an intermediate handler between the :py:class:`Runner`
    exposed to the outside and the :py:class:`EvaluatedStructureFunction`
    which compute the actual observable.

    Parameters
    ----------
        obs_name : ObservableName
            name
        runner : yadism.runner.Runner
            parent reference
    """

    def __init__(self, obs_name, runner):
        # internal managers
        self.obs_name = obs_name
        self.runner = runner
        self.exss = []
        self.cache = {}
        logger.debug("Init %s", self)

    def __repr__(self):
        return self.obs_name.name

    def __len__(self):
        return len(self.exss)

    def load(self, kinematic_configs):
        """
        Loads all kinematic configurations from the run card.

        Parameters
        ----------
            kinematic_configs : list(dict)
                run card input
        """
        self.exss = []
        for kinematics in kinematic_configs:
            self.exss.append(exs.EvaluatedCrossSection(self, kinematics))

    def get_esf(self, obs_name, kin):
        return self.runner.get_sf(obs_name).get_esf(obs_name, kin, use_raw=False)

    def iterate_result(self):
        for exs in self.exss:
            yield exs.get_result()

    def get_result(self):
        """
        Collects the results from all childrens.

        Returns
        -------
            output : list(ESFResult)
                all children outputs
        """
        return list(self.iterate_result())



# -*- coding: utf-8 -*-

from eko import constants

from ..intrinsic import fl_nc
from ..partonic_channel import RSL
from . import partonic_channel as pc


class AsyGluonVV(pc.PartonicChannelAsy):
    def NLO(self):
        def cg(z, args):
            return constants.TR * (16 * z * (1 - z))

        return RSL(cg)


class AsyGluonAA(AsyGluonVV):
    pass


class MatchingIntrinsicSplus(pc.FMatchingQuark):
    ffns = fl_nc.Splus


class MatchingIntrinsicSminus(pc.FMatchingQuark):
    ffns = fl_nc.Sminus


class MatchingGluonSplus(pc.FMatchingGluon):
    ffns = fl_nc.Splus


class MatchingGluonSminus(pc.FMatchingGluon):
    ffns = fl_nc.Sminus



# -*- coding: utf-8 -*-

from ..intrinsic import fl_cc as intrinsic
from ..light import fl_cc as light
from ..partonic_channel import RSL
from . import partonic_channel as pc


class AsyQuark(pc.PartonicChannelAsy, light.NonSinglet):
    def NNLO(self):
        # silence NNLO since heavy NNLO still not available
        return RSL()


class AsyGluon(pc.PartonicChannelAsy):
    def NLO(self):
        as_norm = 2.0

        def reg(z, _args):
            return 4.0 * z * (1.0 - z) * as_norm

        return RSL(reg)


class MatchingIntrinsicSplus(pc.FMatchingQuarkCC):
    ffns = intrinsic.Splus


class MatchingGluonSplus(pc.FMatchingGluonCC):
    ffns = intrinsic.Splus



# -*- coding: utf-8 -*-
"""
This module contains the main loop for the DIS calculations.

There are two ways of using ``yadism``:

* ``Runner``: this class provides a runner that get the *theory* and
  *observables* descriptions as input and manage the whole observables'
  calculation process
* ``run_dis``: a function that wraps the construction of a ``Runner`` object
  and calls the proper method to get the requested output

.. todo::
    decide about ``run_dis`` and document it properly in module header
"""

import copy
import inspect
import io
import logging
import time

import rich
import rich.align
import rich.box
import rich.console
import rich.markdown
import rich.panel
import rich.progress
from eko import basis_rotation as br
from eko import thresholds
from eko.interpolation import InterpolatorDispatcher

from . import log, observable_name
from .coefficient_functions.coupling_constants import CouplingConstants
from .esf import scale_variations as sv
from .input import compatibility, inspector
from .output import Output
from .sf import StructureFunction as SF
from .xs import CrossSection as XS

log.setup()
logger = logging.getLogger(__name__)


class Runner:
    """
    Wrapper to compute a process

    Parameters
    ----------
    theory : dict
        Dictionary with the theory parameters for the evolution (currently
        including PDFSet and DIS process indication).
    observables : dict
        DIS parameters: process description, kinematic specification for the
        requested output.

    Notes
    -----
    For a full description of the content of `theory` and `dis_observables`
    dictionaries read ??.

    .. todo::
        * reference on theory template
        * detailed description of dis_observables entries

    """

    banner = rich.align.Align(
        rich.panel.Panel.fit(
            inspect.cleandoc(
                r"""  __     __       _ _
                     \ \   / /      | (_)
                      \ \_/ /_ _  __| |_ ___ _ __ ___
                       \   / _` |/ _` | / __| '_ ` _ \
                        | | (_| | (_| | \__ \ | | | | |
                        |_|\__,_|\__,_|_|___/_| |_| |_|
                """
            ),
            rich.box.SQUARE,
            padding=1,
            style="magenta",
        ),
        "center",
    )

    def __init__(self, theory: dict, observables: dict):
        # Validate inputs and improve if necessary
        insp = inspector.Inspector(theory, observables)
        insp.perform_all_checks()
        new_theory, new_observables = compatibility.update(theory, observables)

        # Store inputs
        self._theory = new_theory
        self._observables = new_observables

        # Setup eko stuffs
        interpolator = InterpolatorDispatcher.from_dict(self._observables, mode_N=False)

        # Non-eko theory
        coupling_constants = CouplingConstants.from_dict(theory, self._observables)
        pto = theory["PTO"]
        sv_manager = sv.ScaleVariations(
            order=pto,
            interpolator=interpolator,
            activate_ren=new_theory["RenScaleVar"],
            activate_fact=new_theory["FactScaleVar"],
        )

        # Initialize structure functions
        self.managers = dict(
            interpolator=interpolator,
            threshold=thresholds.ThresholdsAtlas.from_dict(new_theory, "kDIS"),
            coupling_constants=coupling_constants,
            sv_manager=sv_manager,
        )
        # FONLL damping powers
        FONLL_damping = bool(theory["DAMP"])
        if FONLL_damping:
            damping_power = theory.get("DAMPPOWER", 2)  # TODO remove defaults?
            damping_power_c = theory.get("DAMPPOWERCHARM", damping_power)
            damping_power_b = theory.get("DAMPPOWERBOTTOM", damping_power)
            damping_power_t = theory.get("DAMPPOWERTOP", damping_power)
            damping_powers = [damping_power_c, damping_power_b, damping_power_t]
        else:
            damping_powers = [2] * 3
        # pass theory params
        intrinsic_range = []
        if theory["IC"] == 1:
            intrinsic_range.append(4)
        self.theory_params = dict(
            pto=pto,
            scheme=theory["FNS"],
            nf_ff=theory["NfFF"],
            intrinsic_range=intrinsic_range,
            m2hq=(theory["mc"] ** 2, theory["mb"] ** 2, theory["mt"] ** 2),
            TMC=theory["TMC"],
            target=new_observables["TargetDIS"],
            GF=theory["GF"],
            M2W=theory["MW"] ** 2,
            M2target=theory["MP"] ** 2,
            FONLL_damping=FONLL_damping,
            damping_powers=damping_powers,
        )
        logger.info("PTO: %d, process: %s", theory["PTO"], new_observables["prDIS"])
        logger.info("FNS: %s, NfFF: %d", theory["FNS"], theory["NfFF"])
        logger.info("Intrinsic: %s", intrinsic_range)
        logger.info(
            "projectile: %s, target: {Z: %g, A: %g}",
            new_observables["ProjectileDIS"],
            *new_observables["TargetDIS"].values(),
        )

        self.observable_instances = {}
        for obs_name, kins in self._observables["observables"].items():
            on = observable_name.ObservableName(obs_name)
            if on.kind in observable_name.xs:
                obs = XS(on, self)
            else:
                # TODO use get_sf?
                obs = SF(on, self)
            # read kinematics
            obs.load(kins)
            self.observable_instances[obs_name] = obs

        # output console
        if log.silent_mode:
            file = io.StringIO()
        else:
            file = None
        self.console = rich.console.Console(file=file)
        # ==============================
        # Initialize output
        # ==============================
        self._output = Output()
        self._output.theory = theory
        self._output.observables = observables
        self._output.update(interpolator.to_dict())
        self._output["pids"] = br.flavor_basis_pids
        self._output["projectilePID"] = coupling_constants.obs_config["projectilePID"]

    def get_sf(self, obs_name):
        if obs_name.name not in self.observable_instances:
            self.observable_instances[obs_name.name] = SF(obs_name, self)
        return self.observable_instances[obs_name.name]

    def get_result(self):
        """
        Compute coefficient functions grid for requested kinematic points.

        Returns
        -------
        :obj:`Output`
            output object, it will store the coefficient functions grid
            (flavour, interpolation-index) for each requested kinematic
            point (x, Q2)

        """
        self.console.print(self.banner)

        # precomputing the plan of calculation
        precomputed_plan = {}
        printable_plan = []
        for name, obs in self.observable_instances.items():
            if name in self._observables["observables"].keys():
                precomputed_plan[name] = obs
                printable_plan.append(
                    f"- {name} at {len(self._observables['observables'][name])} pts"
                )

        self.console.print(rich.markdown.Markdown("## Plan"))
        self.console.print(rich.markdown.Markdown("\n".join(printable_plan)))

        # running the calculation
        self.console.print(rich.markdown.Markdown("## Calculation"))
        self.console.print("yadism took off! please stay tuned ...")
        start = time.time()
        if log.debug:
            for name, obs in precomputed_plan.items():
                self._output[name] = obs.get_result()
        else:
            with rich.progress.Progress(
                transient=True, console=self.console
            ) as progress:
                task = progress.add_task(
                    "Starting...",
                    total=sum([len(obs) for obs in precomputed_plan.values()]),
                )
                for name, obs in precomputed_plan.items():
                    results = []
                    for res in obs.iterate_result():
                        results.append(res)
                        progress.update(
                            task,
                            description=f"Computing [bold green]{name}",
                            advance=1,
                        )
                    self._output[name] = results
        end = time.time()
        diff = end - start
        self.console.print(f"[cyan]took {diff:.2f} s")

        out = copy.deepcopy(self._output)
        return out



# -*- coding: utf-8 -*-
from .. import partonic_channel as pc
from ..partonic_channel import RSL
from . import f3_nc, nnlo


class NonSinglet(f3_nc.NonSinglet):
    def NNLO(self):
        """
        |ref| implements :eqref:`2.9`, :cite:`vogt-f2lcc`.
        """

        return RSL(
            nnlo.xc3ns2p.c3nm2a, nnlo.xc3ns2p.c3ns2b, nnlo.xc3ns2p.c3nm2c, [self.nf]
        )


class Gluon(pc.EmptyPartonicChannel):
    pass


class Singlet(pc.EmptyPartonicChannel):
    pass



# -*- coding: utf-8 -*-
import numpy as np

from ..partonic_channel import RSL
from . import partonic_channel as pc


class Splus(pc.NeutralCurrentBase):
    def LO(self):
        factor = (self.delta / self.ESF.Q2 - self.sigma_pp / self.delta) * self.eta
        return RSL.from_delta(factor)

    def NLO(self):
        return self.mkNLO("l", "splus")


class Sminus(pc.NeutralCurrentBase):
    def LO(self):
        factor = 2.0 * np.sqrt(self.m1sq * self.m2sq) / self.delta * self.eta
        return RSL.from_delta(factor)

    def NLO(self):
        return self.mkNLO("l", "sminus")



# -*- coding: utf-8 -*-
import numpy as np

from .. import splitting_functions as split
from ..partonic_channel import RSL
from . import partonic_channel as pc


class NonSinglet(pc.ChargedCurrentNonSinglet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sf_prefactor = self.labda

    def NLO(self):
        """
        |ref| implements :eqref:`A.1-4` with Table 1, :cite:`gluck-ccheavy`.
        """
        a = 0
        b1 = lambda z: (-1 - z ** 2) * self.sf_prefactor
        b2 = lambda z: (1 - z) * self.sf_prefactor

        return self.h_q(a, b1, b2)


class Gluon(pc.ChargedCurrentGluon):
    def NLO(self):
        """
        |ref| implements :eqref:`A.5-7` with Table 2, :cite:`gluck-ccheavy`.
        """
        as_norm = 2.0

        def reg(z, _args):
            c1 = 2.0 * (1.0 - self.labda)
            c2 = 0
            c3 = -2.0 * (1.0 - z)
            c4 = 2
            return (
                (
                    (
                        split.lo.pqg_single(z, np.array([], dtype=float))
                        / 2.0
                        * (-self.l_labda(z) - np.log(self.labda))
                    )
                    + self.h_g(z, [c1, c2, c3, c4])
                )
                * self.labda
                * as_norm
            )

        return RSL(reg)



# -*- coding: utf-8 -*-
import json
import pathlib

import numpy as np
import pandas as pd
import yaml
from eko import strong_coupling

from . import __version__
from . import observable_name as on
from .esf.result import ESFResult
from .input import compatibility


class MaskedPDF:
    """
    Mask some pids of a PDF set to be 0.

    Parameters
    ----------
        lhapdf_like : callable
            object that provides an xfxQ2 callable (as `lhapdf <https://lhapdf.hepforge.org/>`_
            and :class:`ekomark.toyLH.toyPDF` do) (and thus is in flavor basis)
        active_pids : list(int)
            active PIDs
    """

    def __init__(self, lhapdf_like, active_pids):
        self.parent = lhapdf_like
        self.active_pids = active_pids

    def __getattr__(self, name):
        return self.parent.__getattribute__(name)

    def xfxQ2(self, pid, x, Q2):
        return self.parent.xfxQ2(pid, x, Q2) if pid in self.active_pids else 0.0


class Output(dict):
    """
    Wrapper for the output to help with application
    to PDFs and dumping to file.
    """

    theory = None
    observables = None

    def apply_pdf(self, lhapdf_like):
        r"""
        Compute all observables for the given PDF.

        Parameters
        ----------
            lhapdf_like : object
                object that provides an xfxQ2 callable (as `lhapdf <https://lhapdf.hepforge.org/>`_
                and :class:`ekomark.toyLH.toyPDF` do) (and thus is in flavor basis)

        Returns
        -------
            ret : PDFOutput
                output dictionary with all structure functions for all x, Q2, result and error
        """
        return self.apply_pdf_theory(lhapdf_like, self.theory)

    def apply_pdf_theory(self, lhapdf_like, theory):
        r"""
        Compute all observables for the given PDF.

        Parameters
        ----------
            lhapdf_like : object
                object that provides an xfxQ2 callable (as `lhapdf <https://lhapdf.hepforge.org/>`_
                and :class:`ekomark.toyLH.toyPDF` do) (and thus is in flavor basis)
            theory : dict
                theory dictionary

        Returns
        -------
            ret : PDFOutput
                output dictionary with all structure functions for all x, Q2, result and error
        """
        new_theory, _ = compatibility.update(theory, dict(TargetDIS="proton"))
        sc = strong_coupling.StrongCoupling.from_dict(new_theory)
        alpha_s = lambda muR: sc.a_s(muR ** 2) * 4.0 * np.pi
        alpha_qed = lambda _muR: theory["alphaqed"]
        return self.apply_pdf_alphas_alphaqed_xir_xif(
            lhapdf_like, alpha_s, alpha_qed, theory["XIR"], theory["XIF"]
        )

    def apply_pdf_alphas_alphaqed_xir_xif(
        self, lhapdf_like, alpha_s, alpha_qed, xiR, xiF
    ):
        r"""
        Compute all observables for the given PDF.

        Parameters
        ----------
            lhapdf_like : object
                object that provides an xfxQ2 callable (as `lhapdf <https://lhapdf.hepforge.org/>`_
                and :class:`ekomark.toyLH.toyPDF` do) (and thus is in flavor basis)
            alpha_s : callable
                alpha_s(muR)
            alpha_qed : callable
                alpha_qed
            xiR : float
                ratio renormalization scale to virtuality (linear!)
            xiF : float
                ratio factorization scale to virtuality (linear!)

        Returns
        -------
            ret : PDFOutput
                output dictionary with all structure functions for all x, Q2, result and error
        """
        # iterate
        ret = PDFOutput()

        xgrid = self["interpolation_xgrid"]

        # dispatch onto result
        for obs in self:
            if not on.ObservableName.is_valid(obs):
                continue
            if self[obs] is None:
                continue
            ret[obs] = []
            for kin in self[obs]:
                ret[obs].append(
                    kin.apply_pdf(
                        lhapdf_like, self["pids"], xgrid, alpha_s, alpha_qed, xiR, xiF
                    )
                )
        return ret

    def get_raw(self):
        """
        Serialize result as dict.

        This maps the original numpy matrices to lists.

        Returns
        -------
            out : dict
                dictionary which will be written on output
        """
        out = {}
        # dump raw elements
        for f in ["interpolation_polynomial_degree", "interpolation_is_log"]:
            out[f] = self[f]
        out["pids"] = list(self["pids"])
        # make raw lists
        for k in ["interpolation_xgrid"]:
            out[k] = self[k].tolist()
        for obs in self:
            if not on.ObservableName.is_valid(obs):
                continue
            if self[obs] is None:
                continue
            out[obs] = []
            for kin in self[obs]:
                out[obs].append(kin.get_raw())
        return out

    def dump_pineappl_to_file(self, filename, obsname):
        """
        Write a pineappl grid to file.

        Parameters
        ----------
            filename : str
                output file name
            obsname : str
                observable to be dumped
        """
        # pylint: disable=no-member, too-many-locals
        if len(self[obsname]) <= 0:
            raise ValueError(f"no ESF {obsname}!")
        import pineappl  # pylint: disable=import-outside-toplevel,import-error

        interpolation_xgrid = self["interpolation_xgrid"]
        # interpolation_is_log = self["interpolation_is_log"]
        interpolation_polynomial_degree = self["interpolation_polynomial_degree"]
        lepton_pid = self["projectilePID"]

        # init pineappl objects
        lumi_entries = [
            pineappl.lumi.LumiEntry([(pid, lepton_pid, 1.0)]) for pid in self["pids"]
        ]
        first_esf_result = self[obsname][0]
        orders = [pineappl.grid.Order(*o) for o in first_esf_result.orders]
        bins = len(self[obsname])
        bin_limits = list(map(float, range(0, bins + 1)))
        # subgrid params
        params = pineappl.subgrid.SubgridParams()
        params.set_reweight(False)
        params.set_x_bins(len(interpolation_xgrid))
        params.set_x_max(interpolation_xgrid[-1])
        params.set_x_min(interpolation_xgrid[0])
        params.set_x_order(interpolation_polynomial_degree)

        grid = pineappl.grid.Grid.create(lumi_entries, orders, bin_limits, params)
        limits = []

        # add each ESF as a bin
        for bin_, obs in enumerate(self[obsname]):
            x = obs.x
            Q2 = obs.Q2

            limits.append((Q2, Q2))
            limits.append((x, x))

            # add all orders
            for o, (v, _e) in obs.orders.items():
                order_index = list(first_esf_result.orders.keys()).index(o)
                prefactor = (
                    ((1.0 / (4.0 * np.pi)) ** o[0])
                    * ((-1.0) ** o[2])
                    * ((-1.0) ** o[3])
                )
                # add for each pid/lumi
                for pid_index, pid_values in enumerate(v):
                    pid_values = prefactor * pid_values
                    # grid is empty? skip
                    if not any(np.array(pid_values) != 0):
                        continue
                    subgrid = pineappl.import_only_subgrid.ImportOnlySubgridV1(
                        pid_values[np.newaxis, :, np.newaxis],
                        [Q2],
                        interpolation_xgrid,
                        [1.0],
                    )
                    grid.set_subgrid(order_index, bin_, pid_index, subgrid)
        # set the correct observables
        normalizations = [1.0] * bins
        remapper = pineappl.bin.BinRemapper(normalizations, limits)
        grid.set_remapper(remapper)

        # set the initial state PDF ids for the grid
        grid.set_key_value("initial_state_1", "2212")
        grid.set_key_value("initial_state_2", str(lepton_pid))
        grid.set_key_value(
            "runcard",
            json.dumps(dict(theory=self.theory, observables=self.observables)),
        )
        grid.set_key_value("yadism_version", __version__)
        grid.set_key_value("lumi_id_types", "pdg_mc_ids")

        # dump file
        grid.optimize()
        grid.write(filename)

    def dump_yaml(self, stream=None):
        """
        Serialize result as YAML.

        Parameters
        ----------
            stream : None or stream
                if given, dump is written on it

        Returns
        -------
            dump : any
                result of dump(output, stream), i.e. a string, if no stream is given or
                Null, if self is written sucessfully to stream
        """
        # TODO explicitly silence yaml
        out = self.get_raw()
        return yaml.dump(out, stream)

    def dump_yaml_to_file(self, filename):
        """
        Writes YAML representation to a file.

        Parameters
        ----------
            filename : string
                target file name

        Returns
        -------
            ret : any
                result of dump(output, stream), i.e. Null if written sucessfully
        """
        with open(filename, "w") as f:
            ret = self.dump_yaml(f)
        return ret

    @classmethod
    def load_yaml(cls, stream):
        """
        Load YAML representation from stream

        Parameters
        ----------
            stream : any
                source stream

        Returns
        -------
            obj : cls
                loaded object
        """
        obj = yaml.safe_load(stream)
        # make list numpy
        for k in ["interpolation_xgrid"]:
            obj[k] = np.array(obj[k])
        for obs in obj:
            if not on.ObservableName.is_valid(obs):
                continue
            if obj[obs] is None:
                continue
            for j, kin in enumerate(obj[obs]):
                obj[obs][j] = ESFResult.from_document(kin)
        return cls(obj)

    @classmethod
    def load_yaml_from_file(cls, filename):
        """
        Load YAML representation from file

        Parameters
        ----------
            filename : string
                source file name

        Returns
        -------
            obj : cls
                loaded object
        """
        obj = None
        with open(filename) as o:
            obj = cls.load_yaml(o)
        return obj


class PDFOutput(Output):
    """
    Wrapper for the PDF output to help with dumping to file.
    """

    def get_raw(self):
        """
        Convert the object into a native Python dictionary

        Returns
        -------
            out : dict
                raw dictionary
        """
        out = {}
        for obs in self:
            if self[obs] is None:
                continue
            out[obs] = []
            for kin in self[obs]:
                out[obs].append({k: float(v) for k, v in kin.items()})
        return out

    @classmethod
    def load_yaml(cls, stream):
        """
        Load the object from YAML.

        Parameters
        ----------
            stream : any
                source stream

        Returns
        -------
            obj : cls
                created object
        """
        obj = yaml.safe_load(stream)
        return cls(obj)

    @property
    def tables(self):
        """
        Convert data into a mapping structure functions -> pandas DataFrame
        """
        tables = {}
        for k, v in self.items():
            tables[k] = pd.DataFrame(v)

        return tables

    def dump_tables_to_file(self, filename):
        """
        Write all tables to file.

        Parameters
        ----------
            filename : str
                output file name
        """
        with open(filename, "w") as f:
            for name, table in self.tables.items():
                f.write("\n".join([name, str(table), "\n"]))

    def dump_tables_to_csv(self, dirname):
        """
        Write all tables to separate csv files.

        Parameters
        ----------
            dirname : str
                output directory name
        """
        dirname = pathlib.Path(dirname)
        dirname.mkdir(exist_ok=True)
        for name, table in self.tables.items():
            filename = dirname / f"{name}.csv"
            table.to_csv(filename)



# -*- coding: utf-8 -*-
"""
Defines the :py:class:`StructureFunction` class.

.. todo::
    refer to the sf-esf overview
"""
import logging

from .esf import esf, tmc

logger = logging.getLogger(__name__)


class StructureFunction:
    """
    Represents an abstract structure function.

    This class acts as an intermediate handler between the :py:class:`Runner`
    exposed to the outside and the :py:class:`EvaluatedStructureFunction`
    which compute the actual observable.

    Parameters
    ----------
        obs_name : ObservableName
            name
        runner : yadism.runner.Runner
            parent reference
    """

    def __init__(self, obs_name, runner):
        # internal managers
        self.obs_name = obs_name
        self.runner = runner
        self.esfs = []
        self.cache = {}
        logger.debug("Init %s", self)

    def __getattribute__(self, name):
        runner = object.__getattribute__(self, "runner")
        if name in runner.managers:
            return runner.managers[name]
        if name in runner.theory_params:
            return runner.theory_params[name]
        return super().__getattribute__(name)

    def __repr__(self):
        return self.obs_name.name

    def __len__(self):
        return len(self.esfs)

    def load(self, kinematic_configs):
        """
        Loads all kinematic configurations from the run card.

        Parameters
        ----------
            kinematic_configs : list(dict)
                run card input
        """
        self.esfs = []
        # iterate F* configurations
        for kinematics in kinematic_configs:
            self.esfs.append(self.get_esf(self.obs_name, kinematics, use_raw=False))

    def get_esf(self, obs_name, kinematics, *args, use_raw=True, force_local=False):
        """
        Returns a :py:class:`EvaluatedStructureFunction` instance.

        This wrappers allows

        - TMC to to access raw computations
        - heavy quark matching schemes to access their light counter parts

        It also implements an internal caching system, to speed up the integrals
        in TMC.

        Parameters
        ----------
            obs_name : .observable_name.ObservableName
                structure function name
            kinematics : dict
                kinematic configuration
            args : any
                further arguments passed down to the instance
            use_raw : bool
                eventually use the ESFTMC? (or just the uncorrected one)

        Returns
        -------
            obj : EvaluatedStructureFunction
                created object
        """
        # TODO rethink and refactor method - only used by TMC
        # TODO remove force_local
        # if force_local is active suppress caching to avoid circular dependecy
        if force_local:
            obj = esf.EvaluatedStructureFunction(self, kinematics)
            return obj
        # else we're happy to cache
        # is it us or do we need to delegate?
        if obs_name == self.obs_name:
            # convert to tuple
            key = list(kinematics.values())
            use_tmc_if_available = not use_raw and self.TMC != 0
            key.append(use_tmc_if_available)
            key = tuple(key)
            # TODO how to incorporate args?
            # search
            try:
                return self.cache[key]
            except KeyError:
                if use_tmc_if_available:
                    obj = tmc.ESFTMCmap[obs_name.kind](self, kinematics)
                else:
                    obj = esf.EvaluatedStructureFunction(self, kinematics, *args)
                self.cache[key] = obj
                return obj
        else:
            # ask our parent (as always)
            return self.runner.get_sf(obs_name).get_esf(obs_name, kinematics, *args)

    def iterate_result(self):
        for esf in self.esfs:
            yield esf.get_result()

    def get_result(self):
        """
        Collects the results from all childrens.

        Returns
        -------
            output : list(ESFResult)
                all children outputs
        """
        return list(self.iterate_result())



# -*- coding: utf-8 -*-
"""
The purpose of this module is to provide a runner that implements the semantics
for the input restrictions defined in the following files:

* ``domains.yaml``: in which all the domains restrictions are defined
* ``cross_constraints.yaml``: in which further restrictions are defined,
    each one involving more than one input field
"""

import logging
import pathlib

import yaml

from . import constraints

here = pathlib.Path(__file__).parent

logger = logging.getLogger(__name__)


# ╔═══════════╗
# ║ Inspector ║
# ╚═══════════╝


class Inspector:
    """Instantiate the runner that goes through all constraints.

    Use :func:`check_domains` and :func:`check_cross_constraints` to run the
    check defined respectively in ``domains.yaml`` and
    ``cross_constraints.yaml``.

    Each **default** applied will issue a specific warning.

    Parameters
    ----------
    theory_runcard : dict
        theory inputs from user to inspect and validate
    observables_runcard : dict
        observables inputs from user to inspect and validate

    """

    def __init__(self, theory_runcard, observables_runcard):
        self.theory = theory_runcard
        self.observables = observables_runcard

        domain_file = here / "domains.yaml"
        with open(domain_file, "r") as file:
            self.domains = yaml.safe_load(file)

        cross_constraints_file = here / "cross_constraints.yaml"
        with open(cross_constraints_file, "r") as file:
            self.cross_constraints = yaml.safe_load(file)

    def check_domains(self):
        """
        Iterate over single field constraints (i.e. domains' definitions) and
        immediately raise an error if any input is found outside the boundaries.

        """

        for dom_def in self.domains:
            # load checker with domain definition
            checker = constraints.type_class_map[dom_def["type"]](**dom_def)

            # check value provided by user
            try:
                # retroeve the checker from available checkers and value from
                # user input, apply the first on the latter
                name = dom_def["known_as"] if "known_as" in dom_def else dom_def["name"]
                checker.check_value(
                    value=self.__getattribute__(dom_def["runcard"])[name]
                )
            except KeyError:
                raise ValueError(
                    f"Missing value for '{dom_def['known_as']}' in the input {dom_def['runcard']} runcard"
                )

    def check_cross_constraints(self):
        """
        Iterate over multiple fields constraints (i.e. cross-constraints) and
        immediately raise an error if any input is found outside the boundaries.

        """
        pass

    # def apply_default(self, missing_yields_error=True):
    # """Apply default for missing required arguments"""

    # self.theory = default_manager(self.theory, missing_yields_error)

    def perform_all_checks(self):
        logger.info("Inspecting runcards...")
        self.check_domains()
        self.check_cross_constraints()
        # self.apply_default()
        logger.info("Inspection completed: success ✓")



# -*- coding: utf-8 -*-
"""
This module provides the base class that define the interface for Structure
Function calculation on a given kinematic point (x, Q2) (that is why they are
called *Evaluated*).
"""

import copy
import logging

import numpy as np
from eko import basis_rotation as br

from .. import coefficient_functions as cf
from . import conv
from . import scale_variations as sv
from .result import ESFResult

logger = logging.getLogger(__name__)


class EvaluatedStructureFunction:
    """
    A specific kinematic point for a specific structure function.

    This class implements the structure for all
    the coefficient functions' providers, for a single kinematic point (x,
    Q2), but all the flavours (singlet, nonsinglet, gluon).

    Since the coefficient functions in general are distributions they are
    provided with an internal representation, that respects the interface
    defined by the class :py:class:`conv.DistributionVec`, and the same
    class is used to perform the convolution with the basis functions (see
    :py:class:`eko.InterpolatorDispatcher`), so the final result will
    consist of an array of dimension 2: one dimension corresponding to the
    interpolation grid, the other to the flavour.

    .. _local-caching:

    .. admonition:: Cache

        A part of the overall caching system is implemented at this level.

        The one implemented here is only a **local, isolated** caching,
        i.e.:

        - the first time the instance is asked for computing the result,
          through the :py:meth:`get_result` method, it registers the result;
        - any following call to the :py:meth:`get_result` method will make
          use of the cached result, and will never recompute it.

        If another instance with the same attributes is asked for the result
        it will recompute it from scratch, because any instance is isolated
        and doesn't keep any reference to the others.

    Parameters
    ----------
        SF : StructureFunction
            the parent :py:class:`StructureFunction` instance, provides an
            interface, holds references to global objects (like managers coming
            from :py:mod:`eko`, e.g. :py:class:`InterpolatorDispatcher`) and
            implements the global caching
        kinematics : dict
            the specific kinematic point as a dict with two elements ('x', 'Q2')

    """

    def __init__(self, SF, kinematics: dict):
        x = kinematics["x"]
        if x > 1 or x <= 0:
            raise ValueError("Kinematics 'x' must be in the range (0,1]")
        if kinematics["Q2"] <= 0:
            raise ValueError("Kinematics 'Q2' must be in the range (0,∞)")
        # check domain
        if x < min(SF.interpolator.xgrid_raw):
            raise ValueError(f"x outside xgrid - cannot convolute starting from x={x}")

        self.sf = SF
        self.x = x
        self.Q2 = kinematics["Q2"]
        self.nf = None
        self.process = SF.runner.managers["coupling_constants"].obs_config["process"]
        self.res = ESFResult(self.x, self.Q2, None)
        self._computed = False
        # select available partonic coefficient functions
        self.orders = list(filter(lambda e: e <= SF.pto, range(2 + 1)))

        logger.debug("Init %s", self)

    def __repr__(self):
        return "%s_%s(x=%f,Q2=%f)" % (self.sf.obs_name, self.process, self.x, self.Q2)

    @property
    def zeros(self):
        return np.zeros((len(br.flavor_basis_pids), len(self.sf.interpolator.xgrid)))

    def compute_local(self):
        """
        Here is where the local caching is actually implemented: if the
        coefficient functions are already computed don't do nothing,
        otherwise call :py:meth:`_compute_component` (checks are per flavour).

        In any case no output is provided, but the result is stored in
        instance's attributes (this method is for internal use).
        """
        # something to do?
        if self._computed:
            return
        cfc = cf.Combiner(self)
        full_orders = [(o, 0, 0, 0) for o in self.orders]
        # prepare scale variations
        sv_manager = self.sf.sv_manager
        if sv_manager is not None:
            full_orders = sv.build_orders(self.sf.pto)
        # init orders with 0
        for o in full_orders:
            self.res.orders[o] = [self.zeros, self.zeros]
        # run
        logger.debug("Compute %s", self)
        # iterate all partonic channels
        for cfe in cfc.collect_elems():
            ker_orders = []
            # compute raw coefficient functions
            for o in self.orders:
                rsl = cfe.coeff[o]()
                if rsl is None:
                    continue
                # compute convolution point
                convolution_point = cfe.coeff.convolution_point()
                val, err = conv.convolute_vector(
                    rsl, self.sf.interpolator, convolution_point
                )
                # add the factor x from the LHS
                val, err = convolution_point * val, convolution_point * err
                partons = np.array(
                    [cfe.partons.get(pid, 0.0) for pid in br.flavor_basis_pids]
                )[:, np.newaxis]
                val = val[np.newaxis, :]
                err = err[np.newaxis, :]
                ker_orders.append(((o, 0, 0, 0), (partons, val, err)))

            # apply scale variations
            # deny factorization scale variation for intrinsic
            if cfe.channel != "intrinsic":
                ker_orders.extend(
                    sv_manager.apply_common_scale_variations(ker_orders, cfc.nf)
                )
                ker_orders.extend(
                    sv_manager.apply_diff_scale_variations(ker_orders, cfc.nf)
                )
            else:
                # deny them even as generated from diff
                ker_orders.extend(
                    filter(
                        lambda e: e[0][3] == 0,
                        sv_manager.apply_diff_scale_variations(ker_orders, cfc.nf),
                    )
                )

            # blow up to flavor space
            for o, (partons, val, err) in ker_orders:
                self.res.orders[o][0] += partons @ val
                self.res.orders[o][1] += np.abs(partons) @ err

        self._computed = True

    def get_result(self):
        """
        Compute actual result

        Returns
        -------
        res : ESFResult
            result
        """
        self.compute_local()

        return copy.deepcopy(self.res)



# -*- coding: utf-8 -*-
from ..intrinsic import f3_nc
from . import partonic_channel as pc


class MatchingIntrinsicRplus(pc.FMatchingQuark):
    ffns = f3_nc.Rplus


class MatchingIntrinsicRminus(pc.FMatchingQuark):
    ffns = f3_nc.Rminus



# -*- coding: utf-8 -*-
import numpy as np
from eko import constants

from ..partonic_channel import RSL
from . import partonic_channel as pc


class GluonVV(pc.NeutralCurrentBase):
    def NLO(self):
        """
        |ref| implements :eqref:`D.2`, :cite:`felix-thesis`.
        """
        CF = constants.CF

        def cg(z, args):
            if self.is_below_threshold(z):
                return 0.0
            # fmt: off
            return  self._FHprefactor * (
                3 * CF / 4
                * (-np.pi * self._rho_p(z) ** 3) / (self._rho(z) * self._rho_q)
                * (2 * self._beta(z) + self._rho(z) * np.log(self._chi(z)))
            ) / z
            # fmt: on

        return RSL(cg)


class GluonAA(GluonVV):
    def NLO(self):
        """
        |ref| implements :eqref:`D.5`, :cite:`felix-thesis`.
        """

        VV = super().NLO()

        def cg(z, args):
            if self.is_below_threshold(z):
                return 0.0
            return VV.reg(z, args) - self._FHprefactor * (
                (np.pi * self._rho_p(z) ** 3 / (2 * self._rho(z) ** 2 * self._rho_q))
                * (
                    2 * self._beta(z) * self._rho(z) * self._rho_q
                    - (
                        self._rho(z) ** 2
                        + (-4 + self._rho(z)) * self._rho(z) * self._rho_q
                        + self._rho_q ** 2
                    )
                    * np.log(self._chi(z))
                )
            )

        return RSL(cg)



# -*- coding: utf-8 -*-
from ..partonic_channel import RSL
from . import fl_nc, nnlo


class NonSinglet(fl_nc.NonSinglet):
    def NNLO(self):
        """
        |ref| implements :eqref:`2.9`, :cite:`vogt-f2lcc`.
        """

        return RSL(
            nnlo.xclns2p.clnc2a, loc=nnlo.xclns2p.clnc2c, args=dict(reg=[self.nf])
        )


class Gluon(fl_nc.Gluon):
    pass


class Singlet(fl_nc.Singlet):
    pass



# -*- coding: utf-8 -*-


import numpy as np
from eko import constants

from .. import splitting_functions as split
from ..intrinsic import f2_nc
from ..partonic_channel import RSL
from . import partonic_channel as pc


class AsyGluonVV(pc.PartonicChannelAsy):
    def NLO(self):
        def cg(z, _args):
            L = self.L
            as_norm = 2.0
            return as_norm * (
                split.lo.pqg_single(z, np.array([], dtype=float))
                * (L + np.log((1.0 - z) / z))
                + 2.0 * constants.TR * (-1.0 + 8.0 * z * (1.0 - z))
            )

        return RSL(cg)


class AsyGluonAA(AsyGluonVV):
    pass


class MatchingIntrinsicSplus(pc.FMatchingQuark):
    ffns = f2_nc.Splus


class MatchingIntrinsicSminus(pc.FMatchingQuark):
    ffns = f2_nc.Sminus


class MatchingGluonSplus(pc.FMatchingGluon):
    ffns = f2_nc.Splus


class MatchingGluonSminus(pc.FMatchingGluon):
    ffns = f2_nc.Sminus



# -*- coding: utf-8 -*-
import numpy as np

hqfl = "cbt"


def update(theory, observables):
    """
    Upgrade the legacy theory and observable runcards with the new settings.

    Parameters
    ----------
        theory : dict
            theory runcard
        observables : dict
            observable runcard

    Returns
    -------
        new_theory : dict
            upgraded theory runcard
        new_obs : dict
            upgraded observable runcard
    """
    new_theory = theory.copy()
    new_obs = observables.copy()
    update_fns(new_theory)
    update_scale_variations(new_theory)
    update_target(new_obs)
    return new_theory, new_obs


def update_scale_variations(theory):
    """
    Extract necessity to compute scale variations.

    Parameters
    ----------
        theory : dict
            theory runcard
    """
    theory["RenScaleVar"] = not np.isclose(theory["XIR"], 1.0)
    theory["FactScaleVar"] = not np.isclose(theory["XIF"], 1.0)


def update_target(obs):
    """
    Map TargetDIS string to a (Z,A) dict.

    Parameters
    ----------
        obs : dict
            observable runcard
    """
    target = obs["TargetDIS"]
    if not isinstance(target, str):
        return
    if target == "proton":
        obs["TargetDIS"] = {"Z": 1.0, "A": 1.0}
    elif target == "neutron":
        obs["TargetDIS"] = {"Z": 0.0, "A": 1.0}
    elif target == "isoscalar":
        obs["TargetDIS"] = {"Z": 1.0, "A": 2.0}
    elif target == "iron":
        obs["TargetDIS"] = {
            "Z": 23.403,
            "A": 49.618,
        }  # Fe=26 and we don't know how these factors got inside APFEL
    elif target == "lead":
        obs["TargetDIS"] = {"Z": 82.0, "A": 208.0}  # from hep-ex/0102049
    else:
        raise ValueError(f"Unknown TargetDIS '{target}'")


def update_fns(theory):
    """
    Shifts kcThr around and add the kDIScThr parameter (likewise for all heavy quarks).

    Parameters
    ----------
        theory : dict
            theory runcard
    """
    fns = theory["FNS"]
    nf = theory["NfFF"]
    if fns not in ["FONLL-A"]:  # = fns in FFNS or ZM-VFNS
        # enforce correct settings moving all thresholds to 0 or oo
        if fns == "FFNS":
            ks = [0] * (nf - 3) + [np.inf] * (6 - nf)
            for k, fl in zip(ks, hqfl):
                theory[f"k{fl}Thr"] = k
        # here there is no difference between DGLAP and DIS
        for fl in hqfl:
            theory[f"kDIS{fl}Thr"] = theory[f"k{fl}Thr"]
    else:
        # keep the old setup for the ZM-VFNS part (above)
        for pid in range(nf + 1, 6 + 1):
            fl = hqfl[pid - 4]
            theory[f"kDIS{fl}Thr"] = theory[f"k{fl}Thr"]
        # for the actual value - keep it or fallback to evolution
        hfl = hqfl[nf - 4]
        if "kDIS{hfl}Thr" not in theory:
            theory[f"kDIS{hfl}Thr"] = theory[f"k{hfl}Thr"]
        # erase all lower thresholds, meaning lower than the interesting one (NfFF)
        for pid in range(4, nf + 1):
            fl = hqfl[pid - 4]
            # we actually need to run in the nf-FNS so even kill that one
            theory[f"k{fl}Thr"] = 0
            if pid < nf:
                theory[f"kDIS{fl}Thr"] = 0



# -*- coding: utf-8 -*-

import numpy as np

from .. import splitting_functions as split
from ..intrinsic import f2_cc as intrinsic
from ..light import f2_cc as light
from ..partonic_channel import RSL
from . import partonic_channel as pc


class AsyQuark(pc.PartonicChannelAsy, light.NonSinglet):
    def NNLO(self):
        # silence NNLO since heavy NNLO still not available
        return RSL()


class AsyGluon(pc.PartonicChannelAsy):
    def NLO(self):
        as_norm = 2.0

        def reg(z, _args):
            L = self.L
            return (
                (split.lo.pqg_single(z, np.array([], dtype=float)) / 2.0)
                * (2.0 * np.log((1.0 - z) / z) + L)
                + 8.0 * z * (1.0 - z)
                - 1.0
            ) * as_norm

        return RSL(reg)


class MatchingIntrinsicSplus(pc.FMatchingQuarkCC):
    ffns = intrinsic.Splus


class MatchingGluonSplus(pc.FMatchingGluonCC):
    ffns = intrinsic.Splus



# -*- coding: utf-8 -*-

fake_kind = "??"
sfs = ["F2", "FL", "F3"]
# xs = ["XSreduced", "XSyreduced"]
xs = ["XSHERANC", "XSHERANCAVG", "XSHERACC", "XSCHORUSCC", "XSNUTEVCC"]
kinds = sfs + xs + [fake_kind]
# external flavors:
heavys = ["charm", "bottom", "top"]
asys = [h + "asy" for h in heavys]
heavylights = [h + "light" for h in heavys]
external_flavors = heavys + ["light", "total"] + asys + heavylights
# internally we allow in addition for the flavor families
flavor_families = ["asy", "heavy"]
flavors = external_flavors + flavor_families


class ObservableName:
    """
    Wrapper to observable names to easy split them into two parts.

    Parameters
    ----------
        name : str
            full observable name
    """

    def __init__(self, name):
        p = name.split("_")
        if len(p) == 1:
            self.kind, self.flavor = p[0], "total"
        elif len(p) == 2:
            self.kind, self.flavor = p[0], p[1]
        else:
            raise ValueError(f"Unknown obsname {name}")
        if self.kind not in kinds:
            raise ValueError(f"Unknown kind {self.kind}")
        if self.flavor not in flavors:
            raise ValueError(f"Unknown flavor {self.flavor}")

    @property
    def name(self):
        """joint name"""
        return self.kind + "_" + self.flavor

    def __eq__(self, other):
        """Test equality of kind and flavor"""
        return self.kind == other.kind and self.flavor == other.flavor

    def apply_kind(self, kind):
        """
        Create new object with given kind and our flavor

        Parameters
        ----------
            kind : str
                new kind

        Returns
        -------
            apply_kind : type(self)
                new kind and our flavor
        """
        return type(self)(kind + "_" + self.flavor)

    def apply_asy(self):
        """
        Computes the asymptotic heavy correspondend.

        Returns
        -------
            apply_asy : type(self)
                asymptotic heavy correspondend
        """
        if self.flavor not in heavys:
            raise ValueError(f"observable is not heavy! [{self}]")
        return self.apply_flavor(self.flavor + "asy")

    def __repr__(self):
        """The full name is the representation"""
        return self.name

    def apply_flavor(self, flavor):
        """
        Create new object with given flavor and our kind

        Parameters
        ----------
            flavor : str
                new flavor

        Returns
        -------
            apply_flavor : type(self)
                our kind and new flavor
        """
        return type(self)(self.kind + "_" + flavor)

    @property
    def is_heavy(self):
        """
        Is it a heavy flavor?

        Returns
        -------
            is_heavy : bool
                is a heavy flavor?
        """
        return self.flavor != "light"

    @property
    def is_raw_heavy(self):
        """Is it a raw heavy flavor? i.e. charm, bottom, or, top"""
        return self.flavor in heavys

    @property
    def is_asy(self):
        """Is it a asymptotic raw heavy flavor? i.e. charmasy, bottomasy, or, topasy"""
        return self.flavor in asys

    @property
    def is_heavylight(self):
        """Is it a heavylight flavor? i.e. charmlight, bottomlight, or, toplight"""
        return self.flavor in heavylights

    @property
    def is_composed(self):
        """Is it a composed flavor? i.e. total"""
        return self.flavor == "total"

    @property
    def flavor_family(self):
        """Abstract flavor family name"""
        if self.is_raw_heavy:
            return "heavy"
        if self.is_asy:
            return "asy"
        if self.is_heavylight:
            return "light"
        return self.flavor

    def apply_flavor_family(self):
        """
        Return name with abstract flavor family name

        Returns
        -------
            apply_flavor_family : type(self)
                new ObservableName
        """
        return self.apply_flavor(self.flavor_family)

    @property
    def hqnumber(self):
        """Heavy quark flavor number"""
        if self.is_asy:
            idx = asys.index(self.flavor)
        elif self.is_heavylight:
            idx = heavylights.index(self.flavor)
        else:
            idx = heavys.index(self.flavor)
        return 4 + idx

    @property
    def raw_flavor(self):
        """underlying raw flavor"""
        if self.flavor == "light":
            return self.flavor
        return heavys[self.hqnumber - 4]

    @classmethod
    def has_heavies(cls, names):
        """
        Are there any heavy objects in names?

        Parameters
        ----------
            names : list(str)
                names to check

        Returns
        -------
            has_heavies : bool
                are there heavy obs in names?
        """
        for n in names:
            if not cls.is_valid(n):
                continue
            o = cls(n)
            if o.is_heavy:
                return True
        return False

    @classmethod
    def has_lights(cls, names):
        """
        Are there any light objects in names?

        Parameters
        ----------
            names : list(str)
                names to check

        Returns
        -------
            has_lights : bool
                are there light obs in names?
        """
        for n in names:
            if not cls.is_valid(n):
                continue
            o = cls(n)
            if o.flavor == "light":
                return True
        return False

    @property
    def mass_label(self):
        """Mass label in the theory runcard"""
        if self.flavor == "light":
            return None
        else:
            return f"m{self.flavor[0]}"

    @classmethod
    def is_valid(cls, name):
        """
        Tests whether the name is a valid observable name

        Returns
        -------
            is_valid : bool
                is valid name?
        """
        try:
            cls(name)
            return True
        except ValueError:
            return False



# -*- coding: utf-8 -*-
from .. import kernels


def import_pc_module(kind, process):
    return kernels.import_local(kind, process, __name__)


def generate(esf, nf):
    """
    Collect the heavy coefficient functions

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nf : int
            number of light flavors

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    kind = esf.sf.obs_name.kind
    pcs = import_pc_module(kind, esf.process)
    ihq = nf + 1
    m2hq = esf.sf.m2hq[ihq - 4]
    if esf.process == "CC":
        w = kernels.cc_weights(
            esf.sf.coupling_constants, esf.Q2, kind, kernels.flavors[nf], nf
        )
        return (
            kernels.Kernel(w["ns"], pcs.NonSinglet(esf, nf, m2hq=m2hq)),
            kernels.Kernel(w["g"], pcs.Gluon(esf, nf, m2hq=m2hq)),
        )
    else:
        weights = nc_weights(esf.sf.coupling_constants, esf.Q2, kind, nf)
        gVV = kernels.Kernel(weights["gVV"], pcs.GluonVV(esf, nf, m2hq=m2hq))
        gAA = kernels.Kernel(weights["gAA"], pcs.GluonAA(esf, nf, m2hq=m2hq))
        return (gVV, gAA)


def nc_weights(coupling_constants, Q2, kind, nf):
    """
    Compute heavy NC weights.

    Parameters
    ----------
        coupling_constants : CouplingConstants
            manager for coupling constants
        Q2 : float
            W virtuality
        kind : str
            structure function kind
        nf : int
            number of light flavors

    Returns
    -------
        weights : dict
            mapping pid -> weight for ns, g and s channel
    """
    ihq = nf + 1
    if kind == "F3":
        # weights = {"qVA": {}}
        #     for q in range(1, ihq):
        #         w = coupling_constants.get_weight(q, Q2, kind)
        #         weights["nsVA"][q] = w
        #         weights["nsVA"][-q] = -w
        return {"gVV": {}, "gAA": {}}
    weight_vv = coupling_constants.get_weight(ihq, Q2, "VV")
    weight_aa = coupling_constants.get_weight(ihq, Q2, "AA")
    return {"gVV": {21: weight_vv}, "gAA": {21: weight_aa}}



# -*- coding: utf-8 -*-

import numpy as np
from eko import constants

from .. import splitting_functions as split
from ..intrinsic import f3_cc as intrinsic
from ..light import f3_cc as light
from ..partonic_channel import RSL
from . import partonic_channel as pc


class AsyQuark(pc.PartonicChannelAsy, light.NonSinglet):
    def NNLO(self):
        # silence NNLO since heavy NNLO still not available
        return RSL()


class AsyGluon(pc.PartonicChannelAsy):
    def NLO(self):
        def reg(z, args):
            return -self.L * split.lo.pqg_single(z, args)

        return RSL(reg)


class MatchingIntrinsicRplus(pc.FMatchingQuarkCC):
    ffns = intrinsic.Rplus


class MatchingGluonRplus(pc.FMatchingGluonCC):
    ffns = intrinsic.Rplus



# -*- coding: utf-8 -*-
import numpy as np

from .. import splitting_functions as split
from ..partonic_channel import RSL
from . import partonic_channel as pc


class NonSinglet(pc.ChargedCurrentNonSinglet):
    def NLO(self):
        """
        |ref| implements :eqref:`A.1-4` with Table 1, :cite:`gluck-ccheavy`.
        """
        a = self.ka
        b1 = lambda z: 2 - 2 * z ** 2 - 2 / z
        b2 = lambda z: 2 / z - 1 - z

        return self.h_q(a, b1, b2)


class Gluon(pc.ChargedCurrentGluon):
    def NLO(self):
        """
        |ref| implements :eqref:`A.5-7` with Table 2, :cite:`gluck-ccheavy`.
        """
        as_norm = 2.0

        def reg(z, _args):
            c1 = (
                12.0 * (1 - self.labda) ** 2 - 18 * (1 - self.labda) + 8
            )  # =12l^2 - 6l +2
            c2 = (1.0 - self.labda) / (1.0 - self.labda * z) - 1.0
            c3 = 6.0 * self.labda
            c4 = -12.0 * self.labda
            return (
                (
                    split.lo.pqg_single(z, np.array([], dtype=float))
                    / 2.0
                    * (self.l_labda(z) - np.log(self.labda))
                )
                + self.h_g(z, [c1, c2, c3, c4])
            ) * as_norm

        return RSL(reg)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['yadism',
 'yadism.coefficient_functions',
 'yadism.coefficient_functions.fonll',
 'yadism.coefficient_functions.heavy',
 'yadism.coefficient_functions.intrinsic',
 'yadism.coefficient_functions.light',
 'yadism.coefficient_functions.light.nlo',
 'yadism.coefficient_functions.light.nnlo',
 'yadism.coefficient_functions.splitting_functions',
 'yadism.coefficient_functions.splitting_functions.nlo',
 'yadism.esf',
 'yadism.input']

package_data = \
{'': ['*']}

install_requires = \
['eko>=0.7.1,<0.8.0',
 'numba>=0.53.1,<0.54.0',
 'numpy>=1.21.0,<2.0.0',
 'pandas>=1.3.0,<2.0.0',
 'rich>=10.6.0,<11.0.0',
 'scipy>=1.7.0,<2.0.0']

extras_require = \
{'docs': ['Sphinx>=4.1.1,<5.0.0',
          'sphinx-rtd-theme>=0.5.2,<0.6.0',
          'recommonmark>=0.7.1,<0.8.0',
          'sphinxcontrib-bibtex>=2.3.0,<3.0.0',
          'sphinxcontrib-details-directive>=0.1.0,<0.2.0',
          'nbsphinx>=0.8.6,<0.9.0']}

setup_kwargs = {
    'name': 'yadism',
    'version': '0.8.1',
    'description': 'Yet Another Deep-Inelastic Scattering Module',
    'long_description': '<p align="center">\n  <a href="https://n3pdf.github.io/yadism/"><img alt="Yadism" src="https://raw.githubusercontent.com/N3PDF/yadism/master/docs/_assets/logo/logo.png" width=600></a>\n</p>\n\n<p align="center">\n  <a href="https://github.com/N3PDF/yadism/actions?query=workflow%3A%22yadism%22"><img alt="Tests" src="https://github.com/N3PDF/yadism/workflows/yadism/badge.svg"></a>\n  <a href=\'https://yadism.readthedocs.io/en/latest/?badge=latest\'><img src=\'https://readthedocs.org/projects/yadism/badge/?version=latest\' alt=\'Documentation Status\' /></a>\n  <a href="https://pypi.org/project/yadism/"><img alt="PyPI" src="https://img.shields.io/pypi/v/yadism"/></a>\n  <a href="https://codecov.io/gh/N3PDF/yadism"><img src="https://codecov.io/gh/N3PDF/yadism/branch/master/graph/badge.svg?token=qgCFyUQ6oG" /></a>\n  <a href="https://www.codefactor.io/repository/github/n3pdf/yadism"><img src="https://www.codefactor.io/repository/github/n3pdf/yadism/badge?s=e5a00668b58574b5b056e1aca01c7b25d2c203f8" alt="CodeFactor" /></a>\n  <a href="https://zenodo.org/badge/latestdoi/219968694"><img src="https://zenodo.org/badge/219968694.svg" alt="DOI"></a>\n</p>\n\n<!--Future Badges\n/github/workflow/status/N3PDF/dis/yadism\n\nuse the ones provided by shields.io:\n- example: https://img.shields.io/github/workflow/status/N3PDF/yadism\n\nnote: in order to make shields.io the repo must be public (or accessible to it in some way)\n\noptional:\n- Release-date (github):\n  - /github/release-date/:user/:repo\n- Downloads:\n  - github: /github/downloads/:user/:repo/total\n  - pypi: /pypi/:period/:packageName\n- License:\n  - pypi-license: /pypi/l/:packageName\n  - github-license: /github/license/:user/:repo\n- Activity:\n  - open-issues (github): /github/issues/:user/:repo\n  - open-pull-requests (github): /github/issues-pr/:user/:repo\n- Code size:\n  - github: /github/languages/code-size/:user/:repo\n-->\n\n## Scope of the project\n\nProvide all necessary tools to compute the DIS structure functions and related objects. This project is linked closely to [EKO](https://github.com/N3PDF/eko).\n\n## Installation\n\nAs a user please use [the released version on PyPI](https://pypi.org/project/yadism/),\nthrough your python package manager, e.g. with `pip`:\n\n```sh\npip install yadism\n```\n\n### Dev\n\nFor development just use the install script and poetry (provided by install\nscript):\n\n```sh\npython install.py\npoetry install\n```\n\n## Development\n\nMembers of the development team should always follow the [contribution\nguidelines](.github/contributing.md), to have a uniform strategy in code\ndevelopment and improve collaboration.\n\n## Contributing or contacting the authors\n\nFor any kind of interaction consider before to read [external contribution\nguidelines](.github/contributing.md#external-contributions), otherwise just send\nan email to the authors:\n\n- [Alessandro Candido](mailto:alessandro.candido@mi.infn.it)\n- [Felix Hekhorn](mailto:felix.hekhorn@mi.infn.it)\n- [Giacomo Magni](mailto:gmagni@nikhef.nl)\n',
    'author': 'Alessandro Candido',
    'author_email': 'alessandro.candido@mi.infn.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://n3pdf.github.io/yadism/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
import numpy as np
from eko import constants

from .. import partonic_channel as pc
from .. import splitting_functions as split
from ..partonic_channel import RSL


class PartonicChannelAsy(pc.PartonicChannel):
    def __init__(self, *args, mu2hq):
        super().__init__(*args)
        self.L = np.log(self.ESF.Q2 / mu2hq)


class PartonicChannelAsyIntrinsic(pc.PartonicChannel):
    def __init__(self, *args, m1sq, m2sq):
        super().__init__(*args)
        self.Q2 = self.ESF.Q2
        self.x = self.ESF.x
        self.m1sq = m1sq
        self.m2sq = m2sq
        self.sigma_pm = self.Q2 + self.m2sq - self.m1sq
        self.delta = self.kinematic_delta(self.m1sq, self.m2sq, -self.Q2)
        self.eta = 2.0 * self.Q2 / (self.sigma_pm + self.delta)

    @staticmethod
    def kinematic_delta(a, b, c):
        return np.sqrt(a ** 2 + b ** 2 + c ** 2 - 2 * (a * b + b * c + c * a))

    def convolution_point(self):
        return self.x / self.eta


class FMatching(PartonicChannelAsyIntrinsic):
    ffns = lambda *_args, m1sq, m2sq: None

    def __init__(self, *args, m1sq, m2sq, mu2hq):
        super().__init__(*args, m1sq=m1sq, m2sq=m2sq)
        self.l = np.log(self.Q2 / mu2hq)

    def obj(self):
        return self.ffns(self.ESF, self.nf, m1sq=self.m1sq, m2sq=self.m2sq)

    def parent_lo_local(self):
        parent_LO = self.obj().LO()
        if parent_LO is None:
            return None
        return parent_LO.args["loc"][0]


class FMatchingQuark(FMatching):
    def NLO(self):
        icl = self.parent_lo_local()
        if icl is None:
            return None
        asnorm = 2.0
        l = self.l

        def sing(z, _args):
            # this coefficient function is *almost* proportional to P_qq
            # i.e. 2CF * (1.0 + z ** 2) / (1.0 - z) is the "bare" P_qq
            return (
                asnorm
                * icl
                * constants.CF
                * ((1.0 + z ** 2) / (1.0 - z) * (l - 2.0 * np.log(1.0 - z) - 1.0))
            )

        # MMa:
        # FortranForm@FullSimplify@Integrate[(1 + z^2)/(1 - z) (l - 2 Log[1 - z] - 1), {z, 0, x}, Assumptions -> {0 < x < 1}] # pylint: disable=line-too-long
        def loc(x, _args):
            return (
                -asnorm
                * icl
                * constants.CF
                * (
                    -(x * (4.0 + l * (2.0 + x))) / 2.0
                    + np.log(1.0 - x)
                    * (-1.0 - 2.0 * l + x * (2.0 + x) + 2.0 * np.log(1.0 - x))
                )
            )

        return RSL(sing=sing, loc=loc)


class FMatchingCC(FMatching):
    ffns = lambda *_args, m1sq: None

    def __init__(self, *args, m1sq, mu2hq):
        super().__init__(*args, m1sq=m1sq, m2sq=0.0, mu2hq=mu2hq)

    def obj(self):
        return self.ffns(self.ESF, self.nf, m1sq=self.m1sq)


class FMatchingQuarkCC(FMatchingCC, FMatchingQuark):
    pass


class FMatchingGluon(FMatching):
    def NLO(self):
        return self.mk_nlo_raw(self.l)

    def mk_nlo_raw(self, l):
        icl = self.parent_lo_local()
        if icl is None:
            return None

        # since as and p_qg appear together there is no need to put an explicit as_norm here
        def reg(z, _args):
            return icl * split.lo.pqg_single(z, np.array([], dtype=float)) * l

        return RSL(reg)


class FMatchingGluonCC(FMatchingCC, FMatchingGluon):
    pass



import numpy as np
from eko import beta

from . import lo, nlo

raw_labels = [lo.raw_labels, nlo.raw_labels]


def empty_gluon(matrices, nf):
    return {
        "S_gq": np.zeros_like(matrices["P_qq_0", nf]),
        "S_gg": np.zeros_like(matrices["P_qq_0", nf]),
    }


def joint_lo(fnc, matrices, nf, add_gluonic=True):
    d = {
        "NS_p": fnc("P_qq_0", matrices, nf),
        "NS_m": fnc("P_qq_0", matrices, nf),
        "NS_v": fnc("P_qq_0", matrices, nf),
        "S_qq": fnc("P_qq_0", matrices, nf),
        "S_qg": fnc("P_qg_0", matrices, nf),
    }
    if add_gluonic:
        d["S_gq"] = fnc("P_gq_0", matrices, nf)
        d["S_gg"] = fnc("P_gg_0", matrices, nf)
    else:
        d.update(empty_gluon(matrices, nf))
    return d


def c110(lab, matrices, nf):
    return matrices[lab, nf]


def c211(lab, matrices, nf):
    # add -beta0 * identity
    # where identity is both in interpolation and channel space
    beta0 = beta.beta_0(nf) * np.eye(matrices[lab, nf].shape[0])
    if lab in ["P_gq_0", "P_qg_0"]:
        beta0 = np.zeros_like(matrices[lab, nf])
    return matrices[lab, nf] - beta0


def c220ns(matrices, nf):
    return c220((("P_qq_0^2",), "P_qq_0"), matrices, nf)


def c220(labs, matrices, nf):
    return 0.5 * (
        sum([matrices[lab, nf] for lab in labs[0]])
        - beta.beta_0(nf) * matrices[labs[1], nf]
    )


def sector_mapping(order, matrices, nf):
    smap = {}
    if order >= 1:
        smap.update({(1, 1, 0): joint_lo(c110, matrices, nf, add_gluonic=False)})
    if order >= 2:
        smap.update(
            {
                (2, 1, 0): {
                    "NS_p": matrices["P_nsp_1", nf],
                    "NS_m": matrices["P_nsm_1", nf],
                    "NS_v": matrices["P_nsm_1", nf],
                    "S_qq": matrices["P_qq_1", nf],
                    "S_qg": matrices["P_qg_1", nf],
                    **empty_gluon(matrices, nf),
                },
                (2, 1, 1): joint_lo(c211, matrices, nf),
                (2, 2, 0): {
                    "NS_p": c220ns(matrices, nf),
                    "NS_m": c220ns(matrices, nf),
                    "NS_v": c220ns(matrices, nf),
                    "S_qq": c220(
                        (("P_qq_0^2", "P_qg_0P_gq_0"), "P_qq_0"), matrices, nf
                    ),
                    "S_qg": c220(
                        (("P_qq_0P_qg_0", "P_qg_0P_gg_0"), "P_qg_0"), matrices, nf
                    ),
                    **empty_gluon(matrices, nf),
                },
            }
        )
    return smap
